@@ -9,19 +9,22 @@ import asyncio
 import concurrent.futures
 import contextlib
 import datetime
-import gc
-import json
+import functools
 import logging
 import os
 import os.path
 import pathlib
-import pickle
-import warnings
 from dataclasses import dataclass
 from functools import partial
 from io import UnsupportedOperation
 from typing import List, Iterator, TypeVar, Any, Optional, Callable, AsyncIterator
 
+from .serializer import (
+    ByteSerializer, TextSerializer,
+    JsonSerializer, PickleSerializer, CompressedPickleSerializer,
+    OrjsonSerializer, CompressedOrjsonSerializer,
+)
+
 
 logger = logging.getLogger(__name__)
 T = TypeVar('T', bound='Upath')
@@ -41,24 +44,17 @@ class FileInfo:
     details: Any   # platform-dependent
 
 
-def nogc(func, *args, **kwargs):
-    isgc = gc.isenabled()
-    if isgc:
-        gc.disable()
-    try:
-        return func(*args, **kwargs)
-    finally:
-        if isgc:
-            gc.enable()
-
-
 def _execute_in_thread_pool(jobs, concurrency: int = None):
     if concurrency is None:
         concurrency = 4
     else:
-        assert 0 <= concurrency <= 16
-        if concurrency < 1:
-            concurrency = 1
+        assert 0 <= concurrency <= 32
+
+    if concurrency <= 1:
+        results = []
+        for f, args, kwargs in jobs:
+            results.append(f(*args, **kwargs))
+        return results
 
     pool = concurrent.futures.ThreadPoolExecutor(concurrency)
     tasks = []
@@ -80,11 +76,6 @@ def _should_update(source: Upath, target: Upath) -> bool:
     # previously.
 
 
-def _deprecate(oldname, newname):
-    msg = f"`{oldname}` is deprecated; please use `{newname}` instead"
-    warnings.warn(msg, DeprecationWarning)
-
-
 class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
     @staticmethod
     def _should_overwrite(source: Upath,
@@ -115,6 +106,56 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                 return False
         return True
 
+    @classmethod
+    def register_read_write_byte_format(cls, serde: ByteSerializer, name: str):
+        def _write(self, data, *, overwrite: bool = False):
+            return self.write_bytes(serde.serialize(data), overwrite=overwrite)
+
+        async def _a_write(self, data, *, overwrite: bool = False):
+            return await asyncio.get_running_loop().run_in_executor(
+                None,
+                functools.partial(_write, overwrite=overwrite), self, data,
+            )
+
+        def _read(self):
+            z = self.read_bytes()
+            return serde.deserialize(z)
+
+        async def _a_read(self):
+            return await asyncio.get_running_loop().run_in_executor(
+                None, _read, self,
+            )
+
+        setattr(cls, f'write_{name}', _write)
+        setattr(cls, f'a_write_{name}', _a_write)
+        setattr(cls, f'read_{name}', _read)
+        setattr(cls, f'a_read_{name}', _a_read)
+
+    @classmethod
+    def register_read_write_text_format(cls, serde: TextSerializer, name: str):
+        def _write(self, data, *, overwrite: bool = False):
+            return self.write_text(serde.serialize(data), overwrite=overwrite)
+
+        async def _a_write(self, data, *, overwrite: bool = False):
+            return await asyncio.get_running_loop().run_in_executor(
+                None,
+                functools.partial(_write, overwrite=overwrite), self, data,
+            )
+
+        def _read(self):
+            z = self.read_text()
+            return serde.deserialize(z)
+
+        async def _a_read(self):
+            return await asyncio.get_running_loop().run_in_executor(
+                None, _read, self,
+            )
+
+        setattr(cls, f'write_{name}', _write)
+        setattr(cls, f'a_write_{name}', _a_write)
+        setattr(cls, f'read_{name}', _read)
+        setattr(cls, f'a_read_{name}', _a_read)
+
     def __init__(self, *pathsegments: str, **kwargs):
         '''`Upath` is the base class for a client to a blob store,
         including local file system as a special case.
@@ -470,11 +511,14 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
 
     @ contextlib.contextmanager
     @ abc.abstractmethod
-    def lock(self, *, wait: float = 60):
+    def lock(self, *, timeout: int = None):
         '''Lock the file pointed to, in order to have exclusive access.
 
-        `wait`: if the lock can't be acquired within *wait* seconds,
-        raise `LockAcquisitionTimeoutError`.
+        `timeout`: if the lock can't be acquired within *timeout* seconds,
+        raise `LockAcquisitionTimeoutError`. Default is waiting for ever.
+        Once a lease is acquired, it will not expire until this contexmanager
+        exits. In other word, this is timeout for the "wait", not for the 
+        lease itself. Actual waiting time may be slightly longer.
 
         This is a "mandatory lock", as opposed to an "advisory lock".
         However, this API does not specify that the locked file
@@ -529,17 +573,17 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         '''
         raise NotImplementedError
 
-    def read_json(self, *, no_gc: bool = True, **kwargs):
-        z = self.read_text(**kwargs)
-        if no_gc:
-            return nogc(json.loads, z)
-        return json.loads(z)
+    # def read_json(self, *, no_gc: bool = True, **kwargs):
+    #     z = self.read_text(**kwargs)
+    #     if no_gc:
+    #         return nogc(json.loads, z)
+    #     return json.loads(z)
 
-    def read_pickle(self, *, no_gc: bool = True):
-        z = self.read_bytes()
-        if no_gc:
-            return nogc(pickle.loads, z)
-        return pickle.loads(z)
+    # def read_pickle(self, *, no_gc: bool = True):
+    #     z = self.read_bytes()
+    #     if no_gc:
+    #         return nogc(pickle.loads, z)
+    #     return pickle.loads(z)
 
     def read_text(self, *, encoding: str = 'utf-8', errors: str = 'strict'):
         # Refer to https://docs.python.org/3/library/functions.html#open
@@ -579,25 +623,29 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
                    target: str,
                    *,
                    concurrency: int = None,
-                   exist_action: str = None,
-                   update_filter: Callable[[Upath, Upath], bool] = None,
                    ) -> T:
         '''Analogous to `rename_file`.
 
         Local upath needs to customize this implementation, because
         it needs to take care to delete empty subdirectories under `self`.
         '''
+        if not self.is_dir():
+            raise FileNotFoundError(self)
+
         target_ = self.parent / target
         if target_ == self:
             return self
 
+        if target_.exists():
+            raise FileExistsError(target_)
+
         def foo():
             for p in self.riterdir():
                 extra = str(p.path.relative_to(self.path))
                 yield (
                     p.rename_file,
                     [(target_ / extra)._path],
-                    {'exist_action': exist_action, 'update_filter': update_filter},
+                    {},
                 )
 
         _ = _execute_in_thread_pool(foo(), concurrency)
@@ -610,20 +658,14 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         self._copy_file(target)
         self.remove_file()
 
-    def rename_file(self: T,
-                    target: str,
-                    *,
-                    exist_action: str = None,
-                    update_filter: Callable[[Upath, Upath], bool] = None,
-                    ) -> Optional[T]:
+    def rename_file(self: T, target: str) -> T:
         '''Rename the current file to `target` in the same store.
 
-        `target` is relative to `self.parent`. For example, if `self`
-        is '/a/b/c/d.txt', then `target='e.txt'` means '/a/b/c/e.txt'.
+        `target` is either absolute or relative to `self.parent`.
+        For example, if `self` is '/a/b/c/d.txt', then
+        `target='e.txt'` means '/a/b/c/e.txt'.
 
-        Return an object pointing to the new path if renaming happened.
-
-        If `self` is not an existing file, raise `FileNotFoundError`.
+        Return an object pointing to the new path.
         '''
         if not self.is_file():
             raise FileNotFoundError(self)
@@ -631,17 +673,7 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         if target_ == self:
             return self
 
-        # TODO: rethink this logic.
-        if target_.is_file():
-            if self._should_overwrite(self, target_,
-                                      exist_action=exist_action,
-                                      update_filter=update_filter):
-                target_.remove_file()
-                self._rename_file(target_)
-                return target_
-            return None  # TODO: rethink
-
-        if target_.is_dir():
+        if target_.exists():
             raise FileExistsError(target_)
 
         self._rename_file(target_)
@@ -727,16 +759,16 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         '''
         raise NotImplementedError
 
-    def write_json(self, data, *, overwrite=False, **kwargs) -> int:
-        return self.write_text(json.dumps(data),
-                               overwrite=overwrite,
-                               **kwargs)
+    # def write_json(self, data, *, overwrite=False, **kwargs) -> int:
+    #     return self.write_text(json.dumps(data),
+    #                            overwrite=overwrite,
+    #                            **kwargs)
 
-    def write_pickle(self, data, *, overwrite=False) -> int:
-        return self.write_bytes(
-            pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL),
-            overwrite=overwrite,
-        )
+    # def write_pickle(self, data, *, overwrite=False) -> int:
+    #     return self.write_bytes(
+    #         pickle.dumps(data, protocol=pickle.HIGHEST_PROTOCOL),
+    #         overwrite=overwrite,
+    #     )
 
     def write_text(self,
                    data: str,
@@ -758,59 +790,32 @@ class Upath(abc.ABC):  # pylint: disable=too-many-public-methods
         for p in self.iterdir():
             yield p
 
+    @ contextlib.asynccontextmanager
+    async def a_lock(self, *, timeout: int = None):
+        # TODO: a naive implementation.
+        with self.lock(timeout=timeout):
+            yield
+
     async def a_riterdir(self: T) -> AsyncIterator[T]:
         # TODO: may need reimplementation.
         for p in self.riterdir():
             yield p
 
-    # deprecated on 7/19
 
-    def isdir(self):
-        _deprecate('isdir', 'is_dir')
-        return self.is_dir()
-
-    def isfile(self):
-        _deprecate('isfile', 'is_file')
-        return self.is_file()
-
-    async def a_isdir(self):
-        _deprecate('a_isdir', 'a_is_dir')
-        return await self.a_is_dir()
-
-    async def a_isfile(self):
-        _deprecate('a_isfile', 'a_is_file')
-        return await self.a_is_file()
-
-    def rmdir(self, *, missing_ok=False, concurrency=None):
-        _deprecate('rmdir', 'remove_dir')
-        n = self.remove_dir(concurrency=concurrency)
-        if n == 0 and not missing_ok:
-            raise FileNotFoundError(self)
-        return n
-
-    def rmfile(self, *, missing_ok=False):
-        _deprecate('rmfile', 'remove_file')
-        n = self.remove_file()
-        if n == 0 and not missing_ok:
-            raise FileNotFoundError(self)
-        return n
-
-    async def a_rmdir(self, *, missing_ok=False, concurrency=None):
-        _deprecate('a_rmdir', 'a_remove_dir')
-        n = await self.a_remove_dir(concurrency=concurrency)
-        if n == 0 and not missing_ok:
-            raise FileNotFoundError(self)
-        return n
-
-    async def a_rmfile(self, *, missing_ok=False):
-        _deprecate('a_rmfile', 'a_remove_file')
-        n = await self.a_remove_file()
-        if n == 0 and not missing_ok:
-            raise FileNotFoundError(self)
-        return n
+# Add methods 'read_json', 'write_json', 'a_read_json', 'a_write_json', etc.
+Upath.register_read_write_text_format(JsonSerializer, 'json')
+Upath.register_read_write_byte_format(PickleSerializer, 'pickle')
+Upath.register_read_write_byte_format(CompressedPickleSerializer, 'pickle_z')
+Upath.register_read_write_byte_format(OrjsonSerializer, 'orjson')
+Upath.register_read_write_byte_format(CompressedOrjsonSerializer, 'orjson_z')
 
 
 def make_a_method(name):
+    '''
+    Create an async method named f'a_{name}' based on
+    the sync method 'name'. The async method has the same
+    interface as the sync one, i.e. they take the same parameters.
+    '''
     async def f(self, *args, **kwargs):
         f = partial(getattr(self, name), *args, **kwargs)
         return await asyncio.get_running_loop().run_in_executor(
@@ -821,6 +826,7 @@ def make_a_method(name):
     return f
 
 
+# Add async methods 'a_copy_dir', 'a_copy_file', etc.
 for m in ('copy_dir', 'copy_file',
           'export_dir', 'export_file',
           'exists',
@@ -828,10 +834,10 @@ for m in ('copy_dir', 'copy_file',
           'import_dir', 'import_file',
           'is_dir', 'is_file',
           'ls',
-          'read_bytes', 'read_json', 'read_pickle', 'read_text',
+          'read_bytes', 'read_text',
           'remove_dir', 'remove_file',
           'rename_dir', 'rename_file',
           'rmrf',
-          'write_bytes', 'write_json', 'write_pickle', 'write_text',
+          'write_bytes', 'write_text',
           ):
     setattr(Upath, f'a_{m}', make_a_method(m))
