@@ -6,6 +6,7 @@ from __future__ import annotations
 
 import contextlib
 import logging
+import os
 from io import BufferedReader, UnsupportedOperation
 from google.oauth2 import service_account  # type: ignore
 from google.cloud import storage  # type: ignore
@@ -93,6 +94,7 @@ class GcpBlobUpath(BlobUpath):
     def _export_file(self, target: Upath):
         if not isinstance(target, LocalUpath):
             return super()._export_file(target)
+        os.makedirs(str(target.parent), exist_ok=True)
         self._blob.download_to_filename(str(target))
 
     def file_info(self):
@@ -119,7 +121,11 @@ class GcpBlobUpath(BlobUpath):
         return self._blob.exists()
 
     @contextlib.contextmanager
-    def lock(self, *, wait=60):
+    def lock(self, *, timeout=None):
+        # References:
+        # https://www.joyfulbikeshedding.com/blog/2021-05-19-robust-distributed-locking-algorithm-based-on-google-cloud-storage.html
+        # https://cloud.google.com/storage/docs/generations-preconditions
+        # https://cloud.google.com/storage/docs/gsutil/addlhelp/ObjectVersioningandConcurrencyControl
         raise NotImplementedError
 
     def read_bytes(self):
@@ -139,9 +145,6 @@ class GcpBlobUpath(BlobUpath):
         except NotFound:
             return 0
 
-    def _rename_file(self, target: GcpBlobUpath):
-        self._bucket.rename_blob(self._blob, target._blob_name)
-
     def riterdir(self):
         prefix = self._blob_name + '/'
         k = len(prefix)
