@@ -1,10 +1,12 @@
 # type: ignore
 
 import asyncio
+import concurrent.futures
+import functools
 import pathlib
 import time
 import pytest
-from upathlib import Upath, LocalUpath
+from upathlib import Upath, LocalUpath, LockAcquisitionTimeoutError
 
 # User runs these tests with a `Upath` object of their subclass,
 # in a path that is safe for testing.
@@ -75,8 +77,8 @@ def test_read_write_rm_navigate(p: Upath):
     with pytest.raises(FileExistsError):
         p1.write_text('abcd')
 
-    p1.write_text('abcd', overwrite=True)
-    assert p1.read_text() == 'abcd'
+    p1.write_json({'data': 'abcd'}, overwrite=True)
+    assert p1.read_json() == {'data': 'abcd'}
 
     p /= 'a'
     assert p._path == f'{init_path}/a'
@@ -107,6 +109,8 @@ def test_read_write_rm_navigate(p: Upath):
     print('p2:', fi2)
     print('')
     assert fi1.mtime < fi2.mtime
+    print('file 1 size:', fi1.size)
+    print('file 2 size:', fi2.size)
     assert fi1.size > fi2.size
 
     assert p3.is_dir()
@@ -142,8 +146,8 @@ async def test_a_read_write_rm_navigate(p: Upath):
     with pytest.raises(FileExistsError):
         await p1.a_write_text('abcd')
 
-    await p1.a_write_text('abcd', overwrite=True)
-    assert await p1.a_read_text() == 'abcd'
+    await p1.a_write_json({'data': 'abcd'}, overwrite=True)
+    assert await p1.a_read_json() == {'data': 'abcd'}
 
     p /= 'a'
     assert p._path == f'{init_path}/a'
@@ -156,11 +160,11 @@ async def test_a_read_write_rm_navigate(p: Upath):
     # from the first file.
 
     p2 = p.joinpath('x.data')
-    await p2.a_write_bytes(b'x')
+    await p2.a_write_orjson_z(['x', 'y', 13])
     p /= '..'
     assert p._path == init_path
     assert p2 == p.joinpath('a', 'x.data')
-    assert await p2.a_read_bytes() == b'x'
+    assert await p2.a_read_orjson_z() == ['x', 'y', 13]
 
     p3 = p / 'a'
 
@@ -176,7 +180,9 @@ async def test_a_read_write_rm_navigate(p: Upath):
     print('p2:', fi2)
     print('')
     assert fi1.mtime < fi2.mtime
-    assert fi1.size > fi2.size
+    print('file 1 size:', fi1.size)
+    print('file 2 size:', fi2.size)
+    assert fi1.size < fi2.size
 
     assert await p3.a_is_dir()
     assert await p3.a_remove_dir() == 1
@@ -313,12 +319,36 @@ async def test_a_rename(p: Upath):
     assert not await (p / 'c').a_exists()
 
 
-def test_lock(p: Upath):
-    pass
+def _access_in_mp(root: Upath, path: str, timeout):
+    p = root / path
+    t0 = time.perf_counter()
+    try:
+        with p.lock(timeout=timeout):
+            return time.perf_counter() - t0
+    except LockAcquisitionTimeoutError:
+        return t0 - time.perf_counter()
 
 
-async def test_a_lock(p: Upath):
-    pass
+def test_lock(p: Upath, timeout=None, wait=3):
+    p.rmrf()
+    pp = p / 'testlock'
+    with pp.lock(timeout=timeout):
+        with concurrent.futures.ProcessPoolExecutor(1) as pool:
+            t = pool.submit(_access_in_mp, p / '/', pp._path, wait)
+            z = t.result()
+            print('mp returned after', z, 'seconds')
+            assert z <= -wait
+
+
+async def test_a_lock(p: Upath, timeout=None, wait=3):
+    await p.a_rmrf()
+    pp = p / 'testlock'
+    async with pp.a_lock(timeout=timeout):
+        with concurrent.futures.ProcessPoolExecutor(1) as pool:
+            ff = functools.partial(_access_in_mp, p / '/', pp._path, wait)
+            z = await asyncio.get_running_loop().run_in_executor(pool, ff)
+            print('mp returned after', z, 'seconds')
+            assert z <= -wait
 
 
 def test_all(p: Upath):
@@ -329,11 +359,9 @@ def test_all(p: Upath):
     test_read_write_rm_navigate(p)
     test_import_export(p)
     test_rename(p)
-    test_lock(p)
 
 
 async def test_all_a(p: Upath):
     await test_a_read_write_rm_navigate(p)
     await test_a_import_export(p)
     await test_a_rename(p)
-    await test_a_lock(p)
